@@ -73,10 +73,11 @@ struct rst_info {
 	 */
 	bool has_old_seccomp_filter;
 
-	bool has_thp_enabled;
-
 	struct rst_rseq *rseqe;
 
+	futex_t shstk_enable;
+	futex_t shstk_unlock;
+
 	void *breakpoint;
 };
 
