@@ -1,14 +1,17 @@
 #!/bin/bash
 set -x -e
 
-CI_PKGS="protobuf-c-compiler libprotobuf-c-dev libaio-dev libgnutls28-dev
+CI_PKGS=(protobuf-c-compiler libprotobuf-c-dev libaio-dev libgnutls28-dev
 		libgnutls30 libprotobuf-dev protobuf-compiler libcap-dev
 		libnl-3-dev gdb bash libnet-dev util-linux asciidoctor
-		libnl-route-3-dev time flake8 libbsd-dev python3-yaml
-		libperl-dev pkg-config python3-future python3-protobuf
-		python3-junit.xml"
+		libnl-route-3-dev time libbsd-dev python3-yaml
+		libperl-dev pkg-config python3-protobuf python3-pip
+		python3-importlib-metadata python3-junit.xml libdrm-dev)
 
-X86_64_PKGS="gcc-multilib"
+X86_64_PKGS=(gcc-multilib)
+
+# Convert from string to array.
+IFS=" " read -r -a ZDTM_OPTS <<< "$ZDTM_OPTS"
 
 UNAME_M=$(uname -m)
 
@@ -46,19 +49,15 @@ ci_prep () {
 	else
 		CC=gcc
 	fi
-	CI_PKGS="$CI_PKGS $CC"
+	CI_PKGS+=("$CC")
 
 	# Do not install x86_64 specific packages on other architectures
 	if [ "$UNAME_M" = "x86_64" ]; then
-		CI_PKGS="$CI_PKGS $X86_64_PKGS"
+		CI_PKGS+=("${X86_64_PKGS[@]}")
 	fi
 
-	scripts/ci/apt-install "$CI_PKGS"
+	scripts/ci/apt-install "${CI_PKGS[@]}"
 	chmod a+x "$HOME"
-
-	# zdtm uses an unversioned python binary to run the tests.
-	# let's point python to python3
-	ln -sf /usr/bin/python3 /usr/bin/python
 }
 
 test_stream() {
@@ -69,9 +68,8 @@ test_stream() {
 	# restorer and eventually close the page read. However, image-streamer expects the
 	# whole image to be read and the image is not reopened, sent twice. These MAP_HUGETLB
 	# test cases will result in EPIPE error at the moment.
-	STREAM_TEST_EXCLUDE="-x maps09 -x maps10"
-	# shellcheck disable=SC2086
-	./test/zdtm.py run --stream -p 2 --keep-going -a $STREAM_TEST_EXCLUDE $ZDTM_OPTS
+	STREAM_TEST_EXCLUDE=(-x maps09 -x maps10)
+	./test/zdtm.py run --stream -p 2 --keep-going -a "${STREAM_TEST_EXCLUDE[@]}" "${ZDTM_OPTS[@]}"
 }
 
 print_header() {
@@ -142,6 +140,11 @@ time make unittest
 
 [ -n "$SKIP_CI_TEST" ] && exit 0
 
+# Umount cpuset in cgroupv1 to make it move to cgroupv2
+if [ -d /sys/fs/cgroup/cpuset ]; then
+	umount /sys/fs/cgroup/cpuset
+fi
+
 ulimit -c unlimited
 
 cgid=$$
@@ -160,21 +163,20 @@ if [ "${COMPAT_TEST}x" = "yx" ] ; then
 	# for 32-bit tests. A better way would involve launching docker..
 	# But it would require making zdtm.py aware of docker and launching
 	# tests inside the CT.
-	INCOMPATIBLE_LIBS="libaio-dev libcap-dev libnl-3-dev libnl-route-3-dev"
-	IA32_PKGS=""
+	INCOMPATIBLE_LIBS=(libaio-dev libcap-dev libnl-3-dev libnl-route-3-dev)
+	IA32_PKGS=()
 	REFUGE=64-refuge
 
 	mkdir "$REFUGE"
-	for i in $INCOMPATIBLE_LIBS ; do
+	for i in "${INCOMPATIBLE_LIBS[@]}" ; do
 		for j in $(dpkg --listfiles "$i" | grep '\.so$') ; do
 			cp "$j" "$REFUGE/"
 		done
-		IA32_PKGS="$IA32_PKGS $i:i386"
+		IA32_PKGS+=("$i:i386")
 	done
-	# shellcheck disable=SC2086
-	apt-get remove $INCOMPATIBLE_LIBS
+	apt-get remove "${INCOMPATIBLE_LIBS[@]}"
 	dpkg --add-architecture i386
-	scripts/ci/apt-install "$IA32_PKGS"
+	scripts/ci/apt-install "${IA32_PKGS[@]}"
 	mkdir -p /usr/lib/x86_64-linux-gnu/
 	mv "$REFUGE"/* /usr/lib/x86_64-linux-gnu/
 fi
@@ -211,15 +213,12 @@ if [ "${STREAM_TEST}" = "1" ]; then
 	exit 0
 fi
 
-# shellcheck disable=SC2086
-./test/zdtm.py run -a -p 2 --keep-going $ZDTM_OPTS
+./test/zdtm.py run -a -p 2 --keep-going "${ZDTM_OPTS[@]}"
 if criu/criu check --feature move_mount_set_group; then
-	# shellcheck disable=SC2086
-	./test/zdtm.py run -a -p 2 --mntns-compat-mode --keep-going $ZDTM_OPTS
+	./test/zdtm.py run -a -p 2 --mntns-compat-mode --keep-going "${ZDTM_OPTS[@]}"
 fi
 
-# shellcheck disable=SC2086
-./test/zdtm.py run -a -p 2 --keep-going --criu-config $ZDTM_OPTS
+./test/zdtm.py run -a -p 2 --keep-going --criu-config "${ZDTM_OPTS[@]}"
 
 # Newer kernels are blocking access to userfaultfd:
 # uffd: Set unprivileged_userfaultfd sysctl knob to 1 if kernel faults must be handled without obtaining CAP_SYS_PTRACE capability
@@ -227,17 +226,14 @@ if [ -e /proc/sys/vm/unprivileged_userfaultfd ]; then
 	echo 1 > /proc/sys/vm/unprivileged_userfaultfd
 fi
 
-LAZY_EXCLUDE="-x maps04 -x cmdlinenv00 -x maps007"
+LAZY_EXCLUDE=(-x maps04 -x cmdlinenv00 -x maps007)
 
 LAZY_TESTS='.*(maps0|uffd-events|lazy-thp|futex|fork).*'
-LAZY_OPTS="-p 2 -T $LAZY_TESTS $LAZY_EXCLUDE $ZDTM_OPTS"
+LAZY_OPTS=(-p 2 -T "$LAZY_TESTS" "${LAZY_EXCLUDE[@]}" "${ZDTM_OPTS[@]}")
 
-# shellcheck disable=SC2086
-./test/zdtm.py run $LAZY_OPTS --lazy-pages
-# shellcheck disable=SC2086
-./test/zdtm.py run $LAZY_OPTS --remote-lazy-pages
-# shellcheck disable=SC2086
-./test/zdtm.py run $LAZY_OPTS --remote-lazy-pages --tls
+./test/zdtm.py run "${LAZY_OPTS[@]}" --lazy-pages
+./test/zdtm.py run "${LAZY_OPTS[@]}" --remote-lazy-pages
+./test/zdtm.py run "${LAZY_OPTS[@]}" --remote-lazy-pages --tls
 
 bash -x ./test/jenkins/criu-fault.sh
 if [ "$UNAME_M" == "x86_64" ]; then
@@ -260,10 +256,15 @@ if [ -n "$TRAVIS" ] || [ -n "$CIRCLECI" ]; then
        # Error (criu/tty.c:1014): tty: Don't have tty to inherit session from, aborting
        make -C test/others/shell-job/ run
 fi
+make -C test/others/criu-ns/ run
+make -C test/others/skip-file-rwx-check/ run
 make -C test/others/rpc/ run
 
 ./test/zdtm.py run -t zdtm/static/env00 --sibling
 
+./test/zdtm.py run -t zdtm/static/maps00 --preload-libfault
+./test/zdtm.py run -t zdtm/static/maps02 --preload-libfault
+
 ./test/zdtm.py run -t zdtm/transition/maps007 --pre 2 --dedup
 ./test/zdtm.py run -t zdtm/transition/maps007 --pre 2 --noauto-dedup
 ./test/zdtm.py run -t zdtm/transition/maps007 --pre 2 --page-server
@@ -288,6 +289,45 @@ ip net add test
 ./test/zdtm.py run -t zdtm/static/env00 -t zdtm/transition/fork -t zdtm/static/ghost_holes00 -t zdtm/static/socket-tcp -t zdtm/static/msgque -k always
 ./test/crit-recode.py
 
+# Rootless tests
+# Check if cap_checkpoint_restore is supported and also if unshare -c is supported.
+#
+# Do not run this test in a container (see https://github.com/checkpoint-restore/criu/issues/2312).
+# Before v6.8-rc1~215^2~6, the kernel currently did not show correct device and
+# inode numbers in /proc/pid/maps for stackable file systems.
+skip=0
+findmnt -no FSTYPE / | grep overlay && {
+	./criu/criu check --feature overlayfs_maps || skip=1
+}
+unshare -c /bin/true || skip=1
+capsh --supports=cap_checkpoint_restore || skip=1
+
+if [ "$skip" == 0 ]; then
+	make -C test/zdtm/ cleanout
+	rm -rf test/dump
+	setcap cap_checkpoint_restore,cap_sys_ptrace+eip criu/criu
+	if [ -d /sys/fs/selinux ] && command -v getenforce &>/dev/null; then
+		# Note: selinux in Enforcing mode prevents us from calling clone3() or writing to ns_last_pid on restore; hence set to Permissive for the test and then set back.
+		selinuxmode=$(getenforce)
+		if [ "$selinuxmode" != "Disabled" ]; then
+			setenforce Permissive
+		fi
+
+	fi
+	# Run it as non-root in a user namespace. Since CAP_CHECKPOINT_RESTORE behaves differently in non-user namespaces (e.g. no access to map_files) this tests that we can dump and restore
+	# under those conditions. Note that the "... && true" part is necessary; we need at least one statement after the tests so that bash can reap zombies in the user namespace,
+	# otherwise it will exec the last statement and get replaced and nobody will be left to reap our zombies.
+	sudo --user=#65534 --group=#65534 unshare -Ucfpm --mount-proc -- bash -c "./test/zdtm.py run -t zdtm/static/maps00 -f h --rootless && true"
+	if [ -d /sys/fs/selinux ] && command -v getenforce &>/dev/null; then
+		if [ "$selinuxmode" != "Disabled" ]; then
+			setenforce "$selinuxmode"
+		fi
+	fi
+	setcap -r criu/criu
+else
+	echo "Skipping unprivileged mode tests"
+fi
+
 # more crit testing
 make -C test/others/crit run
 
@@ -303,6 +343,9 @@ make -C test/others/ns_ext run
 # config file parser and parameter testing
 make -C test/others/config-file run
 
+# action script testing
+make -C test/others/action-script run
+
 # Skip all further tests when running with GCOV=1
 # The one test which currently cannot handle GCOV testing is compel/test
 # Probably because the GCOV Makefile infrastructure does not exist in compel
@@ -310,3 +353,14 @@ make -C test/others/config-file run
 
 # compel testing
 make -C compel/test
+
+# amdgpu and cuda plugin testing
+make amdgpu_plugin
+make -C plugins/amdgpu/ test_topology_remap
+./plugins/amdgpu/test_topology_remap
+
+./test/zdtm.py run -t zdtm/static/maps00 -t zdtm/static/maps02 --criu-plugin cuda
+./test/zdtm.py run -t zdtm/static/maps00 -t zdtm/static/maps02 --criu-plugin amdgpu
+./test/zdtm.py run -t zdtm/static/maps00 -t zdtm/static/maps02 --criu-plugin amdgpu cuda
+
+./test/zdtm.py run -t zdtm/static/sigpending -t zdtm/static/pthread00 --mocked-cuda-checkpoint --fault 138
