@@ -9,7 +9,7 @@ prep
 ./test/zdtm.py run -t zdtm/static/maps00 --fault 3 --report report -f h || fail
 
 # FIXME: fhandles looks broken on btrfs
-cat /proc/self/mountinfo | grep -P "/.* / " | grep -q btrfs || NOBTRFS=$?
+findmnt --noheadings --target . | grep -q btrfs || NOBTRFS=$?
 if [ $NOBTRFS -eq 1 ] ; then
 	./test/zdtm.py run -t zdtm/static/inotify_irmap --fault 128 --pre 2 -f uns || fail
 fi
@@ -39,3 +39,11 @@ fi
 ./test/zdtm.py run -t zdtm/static/fpu03 --fault 134 -f h --norst || fail
 # also check for the main thread corruption
 ./test/zdtm.py run -t zdtm/static/fpu00 --fault 134 -f h --norst || fail
+
+# check dont_use_freeze_cgroup
+./test/zdtm.py run -t zdtm/static/env00 --freezecg zdtm:t --fault 137
+./test/zdtm.py run -t zdtm/static/env00 --freezecg zdtm:t --fault 137 --norst
+
+if ./test/zdtm.py run -t zdtm/static/vfork00 --fault 136 --report report -f h ; then
+	fail
+fi
