@@ -1,10 +1,4 @@
-#!/usr/bin/env python
-from __future__ import (
-    absolute_import,
-    division,
-    print_function,
-    unicode_literals
-)
+#!/usr/bin/env python3
 
 import argparse
 import atexit
@@ -38,9 +32,25 @@ from zdtm.criu_config import criu_config
 # File to store content of streamed images
 STREAMED_IMG_FILE_NAME = "img.criu"
 
+# A library used to preload C functions to simulate
+# cases such as partial read with pread().
+LIBFAULT_PATH = os.path.join(
+    os.path.dirname(os.path.abspath(__file__)),
+    "libfault",
+    "libfault.so"
+)
+
+# A directory that contains the CRIU plugins.
+PLUGINS_DIR = os.path.join(
+    os.path.dirname(os.path.abspath(__file__)),
+    "plugins"
+)
+
 prev_line = None
 uuid = uuid.uuid4()
 
+NON_ROOT_UID = 65534
+
 
 def alarm(*args):
     print("==== ALARM ====")
@@ -70,18 +80,32 @@ tests_root = None
 def clean_tests_root():
     global tests_root
     if tests_root and tests_root[0] == os.getpid():
+        subprocess.call(["./umount2", os.path.join(tests_root[1], "dev")])
+        os.rmdir(os.path.join(tests_root[1], "root/root"))
         os.rmdir(os.path.join(tests_root[1], "root"))
+        os.rmdir(os.path.join(tests_root[1], "dev"))
         os.rmdir(tests_root[1])
 
 
 def make_tests_root():
     global tests_root
     if not tests_root:
-        tests_root = (os.getpid(), tempfile.mkdtemp("", "criu-root-", "/tmp"))
+        tmpdir = os.environ.get("TMPDIR", "/tmp")
+        tests_root = (os.getpid(), tempfile.mkdtemp("", "criu-root-", tmpdir))
         atexit.register(clean_tests_root)
         os.mkdir(os.path.join(tests_root[1], "root"))
-    os.chmod(tests_root[1], 0o777)
-    return os.path.join(tests_root[1], "root")
+        os.mkdir(os.path.join(tests_root[1], "root", "root"))
+        # The current file system can be mounted with nodev, so let's create a
+        # new tmpfs mount for /dev.
+        devpath = os.path.join(tests_root[1], "dev")
+        os.mkdir(devpath)
+        # zdtm wants to create files on this mount. User namespace tests are
+        # running with custom user and group mappings.
+        subprocess.check_call(["mount", "-t", "tmpfs", "criu-test-dev", devpath])
+        os.chmod(devpath, 0o777)
+    os.chmod(tests_root[1], 0o755)
+    os.chmod(os.path.join(tests_root[1], "root"), 0o755)
+    return os.path.join(tests_root[1], "root", "root"), os.path.join(tests_root[1], "dev")
 
 
 # Report generation
@@ -177,15 +201,16 @@ class host_flavor:
 
 class ns_flavor:
     __root_dirs = [
-        "/bin", "/sbin", "/etc", "/lib", "/lib64", "/dev", "/dev/pts",
-        "/dev/net", "/tmp", "/usr", "/proc", "/run"
+        "/bin", "/sbin", "/etc", "/lib", "/lib64", "/dev",
+        "/tmp", "/usr", "/proc", "/run"
     ]
+    __dev_dirs = ["pts", "net"]
 
     def __init__(self, opts):
         self.name = "ns"
         self.ns = True
         self.uns = False
-        self.root = make_tests_root()
+        self.root, self.devpath = make_tests_root()
         self.root_mounted = False
 
     def __copy_one(self, fname):
@@ -231,16 +256,19 @@ class ns_flavor:
             self.__copy_one(lib)
 
     def __mknod(self, name, rdev=None):
-        name = "/dev/" + name
+        tdev = stat.S_IFCHR
         if not rdev:
-            if not os.access(name, os.F_OK):
+            if not os.access(os.path.join("/dev", name), os.F_OK):
                 print("Skipping %s at root" % name)
                 return
             else:
-                rdev = os.stat(name).st_rdev
+                s = os.stat(os.path.join("/dev", name))
+                rdev = s.st_rdev
+                if stat.S_ISBLK(s.st_mode):
+                    tdev = stat.S_IFBLK
 
-        name = self.root + name
-        os.mknod(name, stat.S_IFCHR, rdev)
+        name = os.path.join(self.devpath, name)
+        os.mknod(name, tdev, rdev)
         os.chmod(name, 0o666)
 
     def __construct_root(self):
@@ -251,11 +279,18 @@ class ns_flavor:
         for ldir in ["/bin", "/sbin", "/lib", "/lib64"]:
             os.symlink(".." + ldir, self.root + "/usr" + ldir)
 
+    def __construct_dev(self):
+        for dir in self.__dev_dirs:
+            os.mkdir(os.path.join(self.devpath, dir))
+            os.chmod(os.path.join(self.devpath, dir), 0o755)
         self.__mknod("tty", os.makedev(5, 0))
         self.__mknod("null", os.makedev(1, 3))
         self.__mknod("net/tun")
         self.__mknod("rtc")
         self.__mknod("autofs", os.makedev(10, 235))
+        ext_dev = os.getenv("ZDTM_MNT_EXT_DEV")
+        if ext_dev:
+            self.__mknod(os.path.basename(ext_dev))
 
     def __copy_deps(self, deps):
         for d in deps.split('|'):
@@ -267,7 +302,7 @@ class ns_flavor:
 
     def init(self, l_bins, x_bins):
         subprocess.check_call(
-            ["mount", "--make-slave", "--bind", ".", self.root])
+            ["mount", "--make-private", "--bind", ".", self.root])
         self.root_mounted = True
 
         if not os.access(self.root + "/.constructed", os.F_OK):
@@ -278,6 +313,9 @@ class ns_flavor:
                     self.__construct_root()
                     os.mknod(self.root + "/.constructed", stat.S_IFREG | 0o600)
 
+        if not os.access(self.devpath + "/.constructed", os.F_OK):
+            self.__construct_dev()
+            os.mknod(self.devpath + "/.constructed", stat.S_IFREG | 0o600)
         for b in l_bins:
             self.__copy_libs(b)
         for b in x_bins:
@@ -392,19 +430,20 @@ class test_fail_expected_exc(Exception):
 
 
 class zdtm_test:
-    def __init__(self, name, desc, flavor, freezer):
+    def __init__(self, name, desc, flavor, freezer, rootless):
         self.__name = name
         self.__desc = desc
         self.__freezer = None
+        self.__timeout = int(self.__desc.get('timeout') or 30)
+        self.__rootless = rootless
         self.__make_action('cleanout')
         self.__pid = 0
         self.__flavor = flavor
         self.__freezer = freezer
         self._bins = [name]
-        self._env = {}
+        self._env = {'TMPDIR': os.environ.get('TMPDIR', '/tmp')}
         self._deps = desc.get('deps', [])
         self.auto_reap = True
-        self.__timeout = int(self.__desc.get('timeout') or 30)
 
     def __make_action(self, act, env=None, root=None):
         sys.stdout.flush()  # Not to let make's messages appear before ours
@@ -426,7 +465,7 @@ class zdtm_test:
             preexec_fn=self.__freezer and self.__freezer.attach or None)
         if act == "pid":
             try_run_hook(self, ["--post-start"])
-        if s.wait():
+        if s.wait(timeout=self.__timeout):
             raise test_fail_exc(str(s_args))
 
         if self.__freezer:
@@ -439,6 +478,8 @@ class zdtm_test:
         wait_pid_die(int(self.__pid), self.__name, self.__timeout)
 
     def __add_wperms(self):
+        if os.getuid() != 0:
+            return
         # Add write perms for .out and .pid files
         for b in self._bins:
             p = os.path.dirname(b)
@@ -457,6 +498,9 @@ class zdtm_test:
             env['ZDTM_NOTIFY_FDIN'] = "100"
             env['ZDTM_NOTIFY_FDOUT'] = "101"
 
+        if self.__rootless:
+            env['ZDTM_ROOTLESS'] = "1"
+
         if not test_flag(self.__desc, 'suid'):
             # Numbers should match those in criu
             env['ZDTM_UID'] = "18943"
@@ -469,6 +513,7 @@ class zdtm_test:
         if self.__flavor.ns:
             env['ZDTM_NEWNS'] = "1"
             env['ZDTM_ROOT'] = self.__flavor.root
+            env['ZDTM_DEV'] = self.__flavor.devpath
             env['PATH'] = "/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin"
 
             if self.__flavor.uns:
@@ -504,8 +549,15 @@ class zdtm_test:
         self.__freezer.thaw()
         if self.__pid:
             print("Send the %d signal to  %s" % (sig, self.__pid))
-            os.kill(int(self.__pid), sig)
-            self.gone(sig == signal.SIGKILL)
+            try:
+                os.kill(int(self.__pid), sig)
+            except ProcessLookupError:
+                if sig != signal.SIGKILL:
+                    raise
+                print("The process %s doesn't exist" % self.__pid)
+                self.gone(True)
+            else:
+                self.gone(sig == signal.SIGKILL)
 
         self.__flavor.fini()
 
@@ -569,12 +621,18 @@ class zdtm_test:
         return opts
 
     def getdopts(self):
-        return self.__getcropts() + self.__freezer.getdopts(
-        ) + self.__desc.get('dopts', '').split()
+        opts = self.__getcropts() + self.__freezer.getdopts() + \
+            self.__desc.get('dopts', '').split()
+        if self.__flavor.ns:
+            opts += ["--external", "mnt[/dev]:ZDTM_DEV"]
+        return opts
 
     def getropts(self):
-        return self.__getcropts() + self.__freezer.getropts(
-        ) + self.__desc.get('ropts', '').split()
+        opts = self.__getcropts() + self.__freezer.getropts() + \
+            self.__desc.get('ropts', '').split()
+        if self.__flavor.ns:
+            opts += ["--external", "mnt[ZDTM_DEV]:%s" % self.__flavor.devpath]
+        return opts
 
     def unlink_pidfile(self):
         self.__pid = 0
@@ -618,29 +676,41 @@ class zdtm_test:
                 ["make", "zdtm_ct"], env=dict(os.environ, MAKEFLAGS=""))
         if not os.access("zdtm/lib/libzdtmtst.a", os.F_OK):
             subprocess.check_call(["make", "-C", "zdtm/"])
+        if 'preload_libfault' in opts and opts['preload_libfault']:
+            subprocess.check_call(["make", "-C", "libfault/"])
+
+        subprocess.check_call(["make", '--no-print-directory', "-C", "plugins/", "clean"])
+        if 'criu_plugin' in opts and opts['criu_plugin']:
+            for name in opts['criu_plugin']:
+                subprocess.check_call(["make", '--no-print-directory', "-C", "plugins/", f"{name}_plugin.so"])
+
+        if 'mocked_cuda_checkpoint' in opts and opts['mocked_cuda_checkpoint']:
+            subprocess.check_call(["make", "-C", "cuda-checkpoint/"])
+        if 'rootless' in opts and opts['rootless']:
+            return
         subprocess.check_call(
             ["flock", "zdtm_mount_cgroups.lock", "./zdtm_mount_cgroups", str(uuid)])
 
     @staticmethod
     def cleanup():
+        if 'rootless' in opts and opts['rootless']:
+            return
         subprocess.check_call(
             ["flock", "zdtm_mount_cgroups.lock", "./zdtm_umount_cgroups", str(uuid)])
 
 
 def load_module_from_file(name, path):
-    if sys.version_info[0] == 3 and sys.version_info[1] >= 5:
-        import importlib.util
-        spec = importlib.util.spec_from_file_location(name, path)
-        mod = importlib.util.module_from_spec(spec)
-        spec.loader.exec_module(mod)
-    else:
-        import imp
-        mod = imp.load_source(name, path)
+    import importlib.util
+    spec = importlib.util.spec_from_file_location(name, path)
+    mod = importlib.util.module_from_spec(spec)
+    spec.loader.exec_module(mod)
     return mod
 
 
 class inhfd_test:
-    def __init__(self, name, desc, flavor, freezer):
+    def __init__(self, name, desc, flavor, freezer, rootless):
+        if rootless:
+            raise test_fail_exc("This kind of test does not currently support rootless mode")
         self.__name = os.path.basename(name)
         print("Load %s" % name)
         self.__fdtyp = load_module_from_file(self.__name, name)
@@ -801,8 +871,8 @@ class inhfd_test:
 
 
 class groups_test(zdtm_test):
-    def __init__(self, name, desc, flavor, freezer):
-        zdtm_test.__init__(self, 'zdtm/lib/groups', desc, flavor, freezer)
+    def __init__(self, name, desc, flavor, freezer, rootless):
+        zdtm_test.__init__(self, 'zdtm/lib/groups', desc, flavor, freezer, rootless)
         if flavor.ns:
             self.__real_name = name
             with open(name) as fd:
@@ -814,7 +884,7 @@ class groups_test(zdtm_test):
 
         self._bins += self.__subs
         self._deps += get_test_desc('zdtm/lib/groups')['deps']
-        self._env = {'ZDTM_TESTS': self.__real_name}
+        self._env['ZDTM_TESTS'] = self.__real_name
 
     def __get_start_cmd(self, name):
         tdir = os.path.dirname(name)
@@ -824,7 +894,7 @@ class groups_test(zdtm_test):
         subprocess.check_call(s_args + [tname + '.cleanout'])
         s = subprocess.Popen(s_args + ['--dry-run', tname + '.pid'],
                              stdout=subprocess.PIPE)
-        out, _ = s.communicate()
+        out, _ = s.communicate(timeout=self.__timeout)
         cmd = out.decode().splitlines()[-1].strip()
 
         return 'cd /' + tdir + ' && ' + cmd
@@ -868,15 +938,22 @@ class criu_cli:
             fault=None,
             strace=[],
             preexec=None,
-            nowait=False):
+            preload_libfault=False,
+            nowait=False,
+            timeout=60):
         env = dict(
             os.environ,
-            ASAN_OPTIONS="log_path=asan.log:disable_coredump=0:detect_leaks=0")
+            ASAN_OPTIONS="log_path=asan.log:disable_coredump=0:detect_leaks=0",
+            CRIU_LIBS_DIR=PLUGINS_DIR
+        )
 
         if fault:
             print("Forcing %s fault" % fault)
             env['CRIU_FAULT'] = fault
 
+        if preload_libfault:
+            env['LD_PRELOAD'] = LIBFAULT_PATH
+
         cr = subprocess.Popen(strace +
                               [criu_bin, action, "--no-default-config"] + args,
                               env=env,
@@ -884,7 +961,11 @@ class criu_cli:
                               preexec_fn=preexec)
         if nowait:
             return cr
-        return cr.wait()
+        return cr.wait(timeout=timeout)
+
+    @staticmethod
+    def exit_signal(ret):
+        return ret < 0
 
 
 class criu_rpc_process:
@@ -967,7 +1048,9 @@ class criu_rpc:
             fault=None,
             strace=[],
             preexec=None,
-            nowait=False):
+            preload_libfault=False,
+            nowait=False,
+            timeout=None):
         if fault:
             raise test_fail_exc('RPC and FAULT not supported')
         if strace:
@@ -1004,8 +1087,11 @@ class criu_rpc:
             else:
                 raise test_fail_exc('RPC for %s required' % action)
         except crpc.CRIUExceptionExternal as e:
-            print("Fail", e)
-            ret = -1
+            if e.typ != e.resp_typ:
+                ret = -2
+            else:
+                print("Fail", e)
+                ret = -1
         else:
             ret = 0
 
@@ -1018,6 +1104,10 @@ class criu_rpc:
 
         return ret
 
+    @staticmethod
+    def exit_signal(ret):
+        return ret == -2
+
 
 class criu:
     def __init__(self, opts):
@@ -1039,6 +1129,7 @@ class criu:
         self.__dedup = bool(opts['dedup'])
         self.__mdedup = bool(opts['noauto_dedup'])
         self.__user = bool(opts['user'])
+        self.__rootless = bool(opts['rootless'])
         self.__leave_stopped = bool(opts['stop'])
         self.__stream = bool(opts['stream'])
         self.__show_stats = bool(opts['show_stats'])
@@ -1050,7 +1141,9 @@ class criu:
         self.__criu_bin = opts['criu_bin']
         self.__crit_bin = opts['crit_bin']
         self.__pre_dump_mode = opts['pre_dump_mode']
+        self.__preload_libfault = bool(opts['preload_libfault'])
         self.__mntns_compat_mode = bool(opts['mntns_compat_mode'])
+        self.__cuda_checkpoint = bool(opts['mocked_cuda_checkpoint'])
 
         if opts['rpc']:
             self.__criu = criu_rpc
@@ -1133,11 +1226,17 @@ class criu:
         s_args = ["--log-file", log, "--images-dir", self.__ddir(),
                   "--verbosity=4"] + opts
 
+        if self.__cuda_checkpoint:
+            s_args += [ "--libdir" , os.path.join(os.getcwd(), "..", "plugins", "cuda") ]
+
         with open(os.path.join(self.__ddir(), action + '.cropt'), 'w') as f:
             f.write(' '.join(s_args) + '\n')
 
         print("Run criu " + action)
 
+        if self.__rootless:
+            s_args += ["--unprivileged"]
+
         strace = []
         if self.__sat:
             fname = os.path.join(self.__ddir(), action + '.strace')
@@ -1156,7 +1255,10 @@ class criu:
         if action == "restore":
             preexec = None
         else:
-            preexec = self.__user and self.set_user_id or None
+            if os.getuid():
+                preexec = None
+            else:
+                preexec = self.__user and self.set_user_id or None
 
         __ddir = self.__ddir()
 
@@ -1171,8 +1273,10 @@ class criu:
         with open("/proc/sys/kernel/ns_last_pid") as ns_last_pid_fd:
             ns_last_pid = ns_last_pid_fd.read()
 
+        preload_libfault = self.__preload_libfault and action in ['dump', 'pre-dump', 'restore']
+
         ret = self.__criu.run(action, s_args, self.__criu_bin, self.__fault,
-                              strace, preexec, nowait)
+                              strace, preexec, preload_libfault, nowait)
 
         if nowait:
             os.close(status_fds[1])
@@ -1212,8 +1316,8 @@ class criu:
                     return
             rst_succeeded = os.access(
                 os.path.join(__ddir, "restore-succeeded"), os.F_OK)
-            if self.__test.blocking() or (self.__sat and action == 'restore' and
-                                          rst_succeeded):
+            if (self.__test.blocking() and not self.__criu.exit_signal(ret)) or \
+               (self.__sat and action == 'restore' and rst_succeeded):
                 raise test_fail_expected_exc(action)
             else:
                 raise test_fail_exc("CRIU %s" % action)
@@ -1476,10 +1580,11 @@ class criu:
             except Exception:
                 return False
 
-        return criu_cli.run(
-            "check",
-            ["--no-default-config", "--verbosity=0", "--feature", feature],
-            opts['criu_bin']) == 0
+        args = ["--no-default-config", "-verbosity=0", "--feature", feature]
+        if opts['rootless']:
+            args += ["--unprivileged"]
+
+        return criu_cli.run("check", args, opts['criu_bin']) == 0
 
     @staticmethod
     def available():
@@ -1651,6 +1756,15 @@ def get_visible_state(test):
     return files, maps, mounts
 
 
+def has_vsyscall(maps):
+    vsyscall = u"ffffffffff600000-ffffffffff601000"
+    for i in maps:
+        if vsyscall in i:
+            return i
+
+    return None
+
+
 def check_visible_state(test, state, opts):
     new = get_visible_state(test)
 
@@ -1666,9 +1780,9 @@ def check_visible_state(test, state, opts):
         new_maps = new[1][pid]
         if os.getenv("COMPAT_TEST"):
             # the vsyscall vma isn't unmapped from x32 processes
-            vsyscall = u"ffffffffff600000-ffffffffff601000 r-xp"
-            if vsyscall in new_maps and vsyscall not in old_maps:
-                new_maps.remove(vsyscall)
+            entry = has_vsyscall(new_maps)
+            if entry and has_vsyscall(old_maps) is None:
+                new_maps.remove(entry)
         if old_maps != new_maps:
             print("%s: Old maps lost: %s" % (pid, old_maps - new_maps))
             print("%s: New maps appeared: %s" % (pid, new_maps - old_maps))
@@ -1891,7 +2005,7 @@ def do_run_test(tname, tdesc, flavs, opts):
         if opts['dry_run']:
             continue
         flav = flavors[f](opts)
-        t = tclass(tname, tdesc, flav, fcg)
+        t = tclass(tname, tdesc, flav, fcg, opts['rootless'])
         cr_api = criu(opts)
 
         try:
@@ -1979,12 +2093,20 @@ class Launcher:
                   file=self.__file_report)
             print(u"# ", file=self.__file_report)
             print(u"1.." + str(nr_tests), file=self.__file_report)
-        with open("/proc/sys/kernel/tainted") as taintfd:
-            self.__taint = taintfd.read()
+        self.__taint = self.__read_kernel_tainted()
         if int(self.__taint, 0) != 0:
-            print("The kernel is tainted: %r" % self.__taint)
-            if not opts["ignore_taint"] and os.getenv("ZDTM_IGNORE_TAINT") != '1':
-                raise Exception("The kernel is tainted: %r" % self.__taint)
+            self.__report_kernel_taint("The kernel is tainted: %r" % self.__taint)
+
+    @staticmethod
+    def __read_kernel_tainted():
+        with open("/proc/sys/kernel/tainted") as taintfd:
+            return taintfd.read().strip()
+
+    @staticmethod
+    def __report_kernel_taint(msg):
+        print(msg)
+        if not opts["ignore_taint"] and os.getenv("ZDTM_IGNORE_TAINT") != "1":
+            raise Exception(msg)
 
     def __show_progress(self, msg):
         perc = int(self.__nr * 16 / self.__total)
@@ -2010,11 +2132,12 @@ class Launcher:
         if len(self.__subs) >= self.__max:
             self.wait()
 
-        with open("/proc/sys/kernel/tainted") as taintfd:
-            taint = taintfd.read()
+        taint = self.__read_kernel_tainted()
         if self.__taint != taint:
-            raise Exception("The kernel is tainted: %r (%r)" %
-                            (taint, self.__taint))
+            prev_taint = self.__taint
+            self.__taint = taint
+            self.__report_kernel_taint(
+                "The kernel is tainted: %r (was %r)" % (taint, prev_taint))
 
         '''
         The option --link-remap allows criu to hardlink open files back to the
@@ -2042,7 +2165,8 @@ class Launcher:
               'sat', 'script', 'rpc', 'criu_config', 'lazy_pages', 'join_ns',
               'dedup', 'sbs', 'freezecg', 'user', 'dry_run', 'noauto_dedup',
               'remote_lazy_pages', 'show_stats', 'lazy_migrate', 'stream',
-              'tls', 'criu_bin', 'crit_bin', 'pre_dump_mode', 'mntns_compat_mode')
+              'tls', 'criu_bin', 'crit_bin', 'pre_dump_mode', 'mntns_compat_mode',
+              'rootless', 'preload_libfault', 'mocked_cuda_checkpoint')
         arg = repr((name, desc, flavor, {d: self.__opts[d] for d in nd}))
 
         if self.__use_log:
@@ -2052,8 +2176,14 @@ class Launcher:
             logf = None
             log = None
 
+        if opts['rootless'] and os.getuid() == 0:
+            os.setgid(NON_ROOT_UID)
+            os.setuid(NON_ROOT_UID)
+        env = dict(os.environ, CR_CT_TEST_INFO=arg)
+        if opts['mocked_cuda_checkpoint']:
+            env['PATH'] = os.path.join(os.getcwd(), "cuda-checkpoint") + ":" + env["PATH"]
         sub = subprocess.Popen(["./zdtm_ct", "zdtm.py"],
-                               env=dict(os.environ, CR_CT_TEST_INFO=arg),
+                               env=env,
                                stdout=log,
                                stderr=subprocess.STDOUT,
                                close_fds=True)
@@ -2193,9 +2323,21 @@ def all_tests(opts):
                 continue
             files.append(fp)
     excl = list(map(lambda x: os.path.join(desc['dir'], x), desc['exclude']))
-    tlist = list(filter(
+    tlist = list(sorted(filter(
         lambda x: not x.endswith('.checkskip') and not x.endswith('.hook') and
-        x not in excl, map(lambda x: x.strip(), files)))
+        x not in excl, map(lambda x: x.strip(), files))))
+
+    if opts.get('test_shard_count'):
+        if opts.get('test_shard_index') is None:
+            raise KeyError('--test_shard_count > 0 must come with --test_shard_index')
+        slice_idx = opts['test_shard_index']
+        slices = opts['test_shard_count']
+        if slice_idx >= slices:
+            raise IndexError('--test_shard_index not less than --test_shard_count ({} >= {})'.format(slice_idx, slices))
+        slist = list(tlist[slice_idx::slices])
+        print("We're shard #{} of {}. Running {} of {} tests.\n".format(slice_idx, slices, len(slist), len(tlist)))
+        tlist = slist
+
     return tlist
 
 
@@ -2306,11 +2448,6 @@ def run_tests(opts):
         return
 
     torun = list(torun)
-    if opts['keep_going'] and len(torun) < 2:
-        print(
-            "[WARNING] Option --keep-going is more useful when running multiple tests"
-        )
-        opts['keep_going'] = False
 
     if opts['exclude']:
         excl = re.compile(".*(" + "|".join(opts['exclude']) + ")")
@@ -2353,6 +2490,7 @@ def run_tests(opts):
                 "Specify --criu-image-streamer-dir or modify PATH to provide an alternate location")
                 .format(streamer_dir))
 
+    usernsIsSupported = criu.check("userns")
     launcher = Launcher(opts, len(torun))
     try:
         for t in torun:
@@ -2422,7 +2560,7 @@ def run_tests(opts):
                 run_flavs = set(test_flavs) & set(opts_flavs)
             else:
                 run_flavs = set([test_flavs.pop()])
-            if not criu.check("userns"):
+            if not usernsIsSupported:
                 run_flavs -= set(['uns'])
             if opts['user']:
                 # FIXME -- probably uns will make sense
@@ -2591,6 +2729,10 @@ def set_nr_hugepages(nr):
         with open("/proc/sys/vm/nr_hugepages", "w") as f:
             f.write("{}\n".format(nr))
         return orig_hugepages
+    except PermissionError as err:
+        # EACCES is expected when running as non-root, otherwise re-raise the exception.
+        if err.errno != errno.EACCES or os.getuid() == 0:
+            raise
     except OSError as err:
         if err.errno != errno.EOPNOTSUPP:
             raise
@@ -2664,6 +2806,10 @@ def get_cli_args():
     rp.add_argument("--freezecg", help="Use freeze cgroup (path:state)")
     rp.add_argument("--user", help="Run CRIU as regular user",
                     action='store_true')
+    rp.add_argument(
+        "--rootless",
+        help="Run CRIU rootless (uid!=0) (needs CAP_CHECKPOINT_RESTORE)",
+        action='store_true')
     rp.add_argument("--rpc",
                     help="Run CRIU via RPC rather than CLI",
                     action='store_true')
@@ -2724,6 +2870,19 @@ def get_cli_args():
     rp.add_argument("--mntns-compat-mode",
                     help="Use old compat mounts restore engine",
                     action='store_true')
+    rp.add_argument("--test-shard-index", type=int, default=None,
+                    help="Select tests for a shard <index> (0-based)")
+    rp.add_argument("--test-shard-count", type=int, default=0,
+                    help="Specify how many shards are being run (0=sharding disabled; must be the same for all shards)")
+    rp.add_argument("--preload-libfault", action="store_true", help="Run criu with library preload to simulate special cases")
+    rp.add_argument("--criu-plugin",
+                    help="Run tests with CRIU plugin",
+                    choices=['amdgpu', 'cuda'],
+                    nargs='+',
+                    default=None)
+    rp.add_argument("--mocked-cuda-checkpoint",
+                    action="store_true",
+                    help="Run criu with the cuda plugin and the mocked cuda-checkpoint tool")
 
     lp = sp.add_parser("list", help="List tests")
     lp.set_defaults(action=list_tests)
